from __future__ import annotations

import json
from typing import Union

from argrelay.plugin_delegator.delegator_utils import redirect_to_no_func_error
from argrelay.enum_desc.FuncState import FuncState
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.enum_desc.ReservedPropName import ReservedPropName
from argrelay.enum_desc.SpecialFunc import SpecialFunc
from argrelay.plugin_delegator.AbstractDelegator import (
    AbstractDelegator,
    get_func_id_from_invocation_input,
    get_func_id_from_interp_ctx,
)
from argrelay.plugin_loader.client_invocation_utils import prohibit_unconsumed_args
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.runtime_context.SearchControl import SearchControl
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.schema_config_core_server.EnvelopeCollectionSchema import index_props_
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    instance_data_,
    envelope_payload_,
)
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    search_control_list_,
    func_id_,
)
from argrelay.schema_config_interp.SearchControlSchema import populate_search_control, search_control_desc
from argrelay.schema_response.InvocationInput import InvocationInput

collection_name_container_ipos_ = 1
data_envelope_container_ipos_ = 2


class MetadataDelegator(AbstractDelegator):
    """
    Implements FS_74_69_61_79 get set data envelope.
    """

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:

        class_to_collection_map: dict = self.server_config.class_to_collection_map

        collection_search_control = populate_search_control(
            class_to_collection_map,
            ReservedEnvelopeClass.ClassCollectionMeta.name,
            [
                {"collection": ReservedPropName.collection_name.name},
            ],
        )

        func_envelopes = []

        given_function_envelope = {
            instance_data_: {
                func_id_: SpecialFunc.func_id_get_data_envelopes.name,
                delegator_plugin_instance_id_: self.plugin_instance_id,
                search_control_list_: [
                    collection_search_control,
                ],
            },
            ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
            ReservedPropName.help_hint.name: "Get `data_envelope`-s based on their `index_prop`-s.",
            ReservedPropName.func_state.name: FuncState.fs_alpha.name,
            ReservedPropName.func_id.name: SpecialFunc.func_id_get_data_envelopes.name,
        }
        func_envelopes.append(given_function_envelope)

        return func_envelopes

    def run_search_control(
        self,
        interp_ctx: InterpContext,
        function_data_envelope: dict,
        func_param_container_offset: int,
    ) -> Union[SearchControl, None]:
        """
        Provides `search_control` based on `index_prop` for FS_74_69_61_79 get set data envelope.
        """
        search_control_list: list[SearchControl] = self.extract_search_control_from_function_data_envelope(
            function_data_envelope,
        )
        if func_param_container_offset < len(search_control_list):
            return search_control_list[func_param_container_offset]
        elif func_param_container_offset == len(search_control_list):
            collection_name_container = interp_ctx.envelope_containers[collection_name_container_ipos_]
            collection_name_container.data_envelopes = (
                interp_ctx
                .query_engine
                .query_data_envelopes_for(collection_name_container)
            )
            if len(collection_name_container.data_envelopes) != 1:
                return None
            else:
                index_props: list[str] = collection_name_container.data_envelopes[0][envelope_payload_][index_props_]
                collection_name = collection_name_container.data_envelopes[0][ReservedPropName.collection_name.name]

                keys_to_types_list: list[dict] = []
                for index_prop in index_props:
                    # For metadata: arg name is the same as prop name:
                    keys_to_types_list.append({index_prop: index_prop})

                # Construct `search_control` dynamically:
                data_envelope_search_control_dict = populate_search_control(
                    self.server_config.class_to_collection_map,
                    # TODO: TODO_08_25_32_95: redesign `class_to_collection_map`:
                    #       We specify `collection_name` instead of `class_name` while assuming they always match.
                    collection_name,
                    keys_to_types_list,
                )
                data_envelope_search_control_obj = search_control_desc.dict_schema.load(
                    data_envelope_search_control_dict,
                )
                return data_envelope_search_control_obj

        else:
            return None

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:

        assert interp_ctx.is_func_found(), "the (first) function envelope must be found"

        func_id = get_func_id_from_interp_ctx(interp_ctx)

        vararg_container_ipos = data_envelope_container_ipos_

        if func_id in [
            SpecialFunc.func_id_get_data_envelopes.name,
        ]:
            # Verify that func is selected and all what is left to do is to query 0...N objects:
            if interp_ctx.curr_container_ipos >= vararg_container_ipos:
                # Search `data_envelope`-s based on existing args on command line:
                vararg_container = interp_ctx.envelope_containers[vararg_container_ipos]
                vararg_container.data_envelopes = (
                    local_server
                    .get_query_engine()
                    .query_data_envelopes_for(vararg_container)
                )

                # Plugin to invoke on client side:
                delegator_plugin_instance_id = self.plugin_instance_id
                # Package into `InvocationInput` payload object:
                invocation_input = InvocationInput.with_interp_context(
                    interp_ctx,
                    delegator_plugin_entry = local_server.plugin_config.plugin_instance_entries[
                        delegator_plugin_instance_id
                    ],
                    custom_plugin_data = {},
                )
                return invocation_input
            else:
                return redirect_to_no_func_error(
                    interp_ctx,
                    local_server.server_config,
                )
        else:
            # Plugin is given a function name it does not know:
            raise RuntimeError

    @staticmethod
    def invoke_action(
        invocation_input: InvocationInput,
    ):
        func_id = get_func_id_from_invocation_input(invocation_input)
        if func_id == SpecialFunc.func_id_get_data_envelopes.name:
            prohibit_unconsumed_args(invocation_input)
            for data_envelope in invocation_input.envelope_containers[data_envelope_container_ipos_].data_envelopes:
                print(json.dumps(data_envelope))



"""
"""
from __future__ import annotations

from typing import Callable, Any, Union

from argrelay.enum_desc.ArgSource import ArgSource
from argrelay.plugin_config.ConfiguratorAbstract import ConfiguratorAbstract
from argrelay.plugin_delegator.ErrorDelegator import ErrorDelegator
from argrelay.plugin_delegator.ErrorDelegatorCustomDataSchema import (
    error_message_,
    error_code_,
    error_delegator_custom_data_desc,
)
from argrelay.runtime_data.AssignedValue import AssignedValue
from argrelay.runtime_data.PluginConfig import PluginConfig
from argrelay.schema_response.InvocationInput import InvocationInput


def set_default_to(
    prop_name,
    prop_value,
    envelope_container,
) -> bool:
    """
    Assign a default `prop_value` to the given `prop_name`:
    *   Ensure that this `prop_name` is valid for `envelope_container`.
    *   Ensure that this `prop_value` is actually one of the remaining ones.
    """

    if prop_name in envelope_container.search_control.types_to_keys_dict:
        if prop_name not in envelope_container.assigned_types_to_values:
            if prop_name in envelope_container.remaining_types_to_values:
                if prop_value in envelope_container.remaining_types_to_values[prop_name]:
                    del envelope_container.remaining_types_to_values[prop_name]
                    envelope_container.assigned_types_to_values[prop_name] = AssignedValue(
                        prop_value,
                        ArgSource.DefaultValue,
                    )
                    return True
        else:
            # Setting `ArgSource.DefaultValue`: it cannot override any, right? No point to handle assigned case:
            pass
    return False


def get_config_value_once(
    server_configurators,
    value_getter: Callable[[ConfiguratorAbstract], Any],
    default_value: Any,
) -> Any:
    config_value: Union[Any, None] = None
    server_configurator: ConfiguratorAbstract
    for server_configurator in server_configurators.values():
        if config_value is None:
            config_value = value_getter(server_configurator)
        else:
            if value_getter(server_configurator) is not None:
                # Only one `PluginType.ConfiguratorPlugin` providing
                # same type of value is supported to avoid confusion:
                raise RuntimeError

    if config_value is None:
        config_value = default_value
    return config_value


def redirect_to_error(
    interp_ctx,
    plugin_config: PluginConfig,
    error_message,
    error_code,
):
    # Redirect to `ErrorDelegator`:
    # TODO: TODO_62_75_33_41: Do not hardcode plugin instance id (instance of `ErrorDelegator`):
    delegator_plugin_instance_id = f"{ErrorDelegator.__name__}.default"
    custom_plugin_data = {
        error_message_: error_message,
        error_code_: error_code,
    }
    error_delegator_custom_data_desc.validate_dict(custom_plugin_data)
    invocation_input = InvocationInput.with_interp_context(
        interp_ctx,
        delegator_plugin_entry = plugin_config.plugin_instance_entries[delegator_plugin_instance_id],
        custom_plugin_data = custom_plugin_data,
    )
    return invocation_input


def redirect_to_no_func_error(
    interp_ctx,
    plugin_config,
):
    return redirect_to_error(
        interp_ctx,
        plugin_config,
        "ERROR: objects cannot be searched until function is fully qualified",
        1,
    )



from __future__ import annotations

import json

from argrelay.custom_integ.ServiceEnvelopeClass import ServiceEnvelopeClass
from argrelay.custom_integ.ServicePropName import ServicePropName
from argrelay.custom_integ.value_constants import (
    func_id_goto_host_,
    func_id_goto_service_,
    func_id_list_host_,
    func_id_list_service_,
    func_id_diff_service_,
    func_id_desc_host_,
    func_id_desc_service_,
)
from argrelay.enum_desc.FuncState import FuncState
from argrelay.enum_desc.ReservedEnvelopeClass import ReservedEnvelopeClass
from argrelay.enum_desc.ReservedPropName import ReservedPropName
from argrelay.plugin_delegator.AbstractDelegator import (
    AbstractDelegator,
    get_func_id_from_interp_ctx,
    get_func_id_from_invocation_input,
)
from argrelay.plugin_delegator.ErrorDelegatorCustomDataSchema import (
    error_message_,
    error_code_,
    error_delegator_stub_custom_data_example,
)
from argrelay.plugin_delegator.delegator_utils import set_default_to, redirect_to_error, redirect_to_no_func_error
from argrelay.plugin_loader.client_invocation_utils import prohibit_unconsumed_args
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.runtime_data.ServerConfig import ServerConfig
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    instance_data_,
)
from argrelay.schema_config_interp.FunctionEnvelopeInstanceDataSchema import (
    delegator_plugin_instance_id_,
    search_control_list_,
    func_id_,
)
from argrelay.schema_config_interp.SearchControlSchema import (
    populate_search_control,
)
from argrelay.schema_response.InvocationInput import InvocationInput

host_container_ipos_ = 1
service_container_ipos_ = 1
access_container_ipos_ = 2

diff_service_container_left_ipos_ = 1
diff_service_container_right_ipos_ = 2


class ServiceDelegator(AbstractDelegator):

    def __init__(
        self,
        server_config: ServerConfig,
        plugin_instance_id: str,
        plugin_config_dict: dict,
    ):
        super().__init__(
            server_config,
            plugin_instance_id,
            plugin_config_dict,
        )

    def get_supported_func_envelopes(
        self,
    ) -> list[dict]:

        class_to_collection_map: dict = self.server_config.class_to_collection_map

        cluster_search_control = populate_search_control(
            class_to_collection_map,
            ServiceEnvelopeClass.ClassCluster.name,
            [
                {"code": ServicePropName.code_maturity.name},
                {"stage": ServicePropName.flow_stage.name},
                {"region": ServicePropName.geo_region.name},
                {"cluster": ServicePropName.cluster_name.name},
            ],
        )

        host_search_control = populate_search_control(
            class_to_collection_map,
            ServiceEnvelopeClass.ClassHost.name,
            [
                # ClassCluster:
                {"code": ServicePropName.code_maturity.name},
                {"stage": ServicePropName.flow_stage.name},
                {"region": ServicePropName.geo_region.name},
                {"cluster": ServicePropName.cluster_name.name},
                # ClassHost:
                {"host": ServicePropName.host_name.name},
                # ---
                {"status": ServicePropName.live_status.name},
                {"ip": ServicePropName.ip_address.name},
            ],
        )

        service_search_control = populate_search_control(
            class_to_collection_map,
            ServiceEnvelopeClass.ClassService.name,
            [
                # ClassCluster:
                {"code": ServicePropName.code_maturity.name},
                {"stage": ServicePropName.flow_stage.name},
                {"region": ServicePropName.geo_region.name},
                {"cluster": ServicePropName.cluster_name.name},
                # ClassService:
                {"group": ServicePropName.group_label.name},
                {"service": ServicePropName.service_name.name},
                {"mode": ServicePropName.run_mode.name},
                # ClassHost:
                {"host": ServicePropName.host_name.name},
                # ---
                {"status": ServicePropName.live_status.name},
                {"dc": ServicePropName.data_center.name},
                {"ip": ServicePropName.ip_address.name},
            ],
        )

        access_search_control = populate_search_control(
            class_to_collection_map,
            ServiceEnvelopeClass.ClassAccessType.name,
            [
                {"access": ServicePropName.access_type.name},
            ],
        )

        func_envelopes = [
            {
                instance_data_: {
                    func_id_: func_id_goto_host_,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                        host_search_control,
                        access_search_control,
                    ],
                },
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedPropName.help_hint.name: "Go (log in) to remote host",
                ReservedPropName.func_state.name: FuncState.fs_demo.name,
                ReservedPropName.func_id.name: func_id_goto_host_,
            },
            {
                instance_data_: {
                    func_id_: func_id_goto_service_,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                        service_search_control,
                        access_search_control,
                    ],
                },
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedPropName.help_hint.name: "Go (log in) to remote host and dir path with specified service",
                ReservedPropName.func_state.name: FuncState.fs_demo.name,
                ReservedPropName.func_id.name: func_id_goto_service_,
            },
            {
                instance_data_: {
                    func_id_: func_id_desc_host_,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                        host_search_control,
                    ],
                },
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedPropName.help_hint.name: "Describe remote host",
                ReservedPropName.func_state.name: FuncState.fs_demo.name,
                ReservedPropName.func_id.name: func_id_desc_host_,
            },
            {
                instance_data_: {
                    func_id_: func_id_desc_service_,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                        service_search_control,
                    ],
                },
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedPropName.help_hint.name: "Describe service instance",
                ReservedPropName.func_state.name: FuncState.fs_demo.name,
                ReservedPropName.func_id.name: func_id_desc_service_,
            },
            {
                instance_data_: {
                    func_id_: func_id_list_host_,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                        host_search_control,
                    ],
                },
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedPropName.help_hint.name: "List remote hosts matching search query",
                ReservedPropName.func_state.name: FuncState.fs_demo.name,
                ReservedPropName.func_id.name: func_id_list_host_,
            },
            {
                instance_data_: {
                    func_id_: func_id_list_service_,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                        service_search_control,
                    ],
                },
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedPropName.help_hint.name: "List service instances matching search query",
                ReservedPropName.func_state.name: FuncState.fs_demo.name,
                ReservedPropName.func_id.name: func_id_list_service_,
            },
            {
                instance_data_: {
                    func_id_: func_id_diff_service_,
                    delegator_plugin_instance_id_: self.plugin_instance_id,
                    search_control_list_: [
                        # This function was created to demo FS_97_64_39_94 `arg_bucket`-s:
                        # it intentionally uses two services as to specify in its args:
                        service_search_control,
                        service_search_control,
                    ],
                },
                ReservedPropName.envelope_class.name: ReservedEnvelopeClass.ClassFunction.name,
                ReservedPropName.help_hint.name: "Diff two service instances",
                ReservedPropName.func_state.name: FuncState.fs_demo.name,
                ReservedPropName.func_id.name: func_id_diff_service_,
            },
        ]
        return func_envelopes

    def has_fill_control(
        self,
    ) -> bool:
        return True

    def run_fill_control(
        self,
        interp_ctx: "InterpContext",
    ) -> bool:
        func_id = get_func_id_from_interp_ctx(interp_ctx)
        any_assignment = False

        if func_id in [
            func_id_goto_host_,
            func_id_goto_service_,
        ]:
            assert host_container_ipos_ == service_container_ipos_
            object_container_ipos = host_container_ipos_

            if func_id == func_id_goto_service_:
                any_assignment = self._fill_service_container(
                    any_assignment,
                    interp_ctx,
                    object_container_ipos,
                )

            # If we need to specify `access_type` `data_envelope`:
            if interp_ctx.curr_container_ipos == interp_ctx.curr_interp.base_container_ipos + access_container_ipos_:
                # Take object found so far:
                data_envelope = interp_ctx.envelope_containers[(
                    interp_ctx.curr_interp.base_container_ipos + object_container_ipos
                )].data_envelopes[0]

                access_container = interp_ctx.envelope_containers[(
                    interp_ctx.curr_interp.base_container_ipos + access_container_ipos_
                )]

                # Select default value to search `access_type` `data_envelope` based on `code_maturity`:
                code_arg_type = ServicePropName.code_maturity.name
                if code_arg_type in data_envelope:
                    code_arg_val = data_envelope[code_arg_type]
                    if code_arg_val == "prod":
                        any_assignment = (
                            set_default_to(ServicePropName.access_type.name, "ro", access_container)
                            or
                            any_assignment
                        )
                    else:
                        any_assignment = (
                            set_default_to(ServicePropName.access_type.name, "rw", access_container)
                            or
                            any_assignment
                        )
        elif func_id in [
            func_id_diff_service_,
        ]:
            any_assignment = self._fill_service_container(
                any_assignment,
                interp_ctx,
                diff_service_container_left_ipos_,
            )
            any_assignment = self._fill_service_container(
                any_assignment,
                interp_ctx,
                diff_service_container_right_ipos_,
            )
        elif func_id in [
            func_id_list_host_,
            func_id_list_service_,
        ]:
            pass
        elif func_id in [
            func_id_desc_host_,
            func_id_desc_service_,
        ]:
            pass
        else:
            raise RuntimeError

        return any_assignment

    def _fill_service_container(
        self,
        any_assignment,
        interp_ctx,
        object_container_ipos,
    ):
        if (
            interp_ctx.curr_container_ipos == interp_ctx.curr_interp.base_container_ipos + object_container_ipos
        ):
            service_container = interp_ctx.envelope_containers[(
                interp_ctx.curr_interp.base_container_ipos + object_container_ipos
            )]
            any_assignment = (
                set_default_to(ServicePropName.run_mode.name, "active", service_container)
                or
                any_assignment
            )
        return any_assignment

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        assert interp_ctx.is_func_found(), "the (first) function envelope must be found"

        func_id = get_func_id_from_interp_ctx(interp_ctx)

        vararg_container_ipos = host_container_ipos_
        assert vararg_container_ipos == host_container_ipos_ == service_container_ipos_

        if func_id in [
            func_id_goto_host_,
            func_id_goto_service_,
        ]:
            # Even if these functions do not support varargs, when `redirect_to_error`, query all:
            vararg_container = interp_ctx.envelope_containers[vararg_container_ipos]
            vararg_container.data_envelopes = (
                local_server
                .get_query_engine()
                .query_data_envelopes_for(vararg_container)
            )

            # Actual implementation is not defined for demo:
            return redirect_to_error(
                interp_ctx,
                local_server.plugin_config,
                error_delegator_stub_custom_data_example[error_message_],
                error_delegator_stub_custom_data_example[error_code_],
            )
        elif func_id in [
            func_id_diff_service_,
        ]:
            # TODO: TODO_75_52_01_67: `arg_bucket`-s to support multiple var args:
            #       query both service lists and compare them.

            # Actual implementation is not defined for demo:
            return redirect_to_error(
                interp_ctx,
                local_server.plugin_config,
                error_delegator_stub_custom_data_example[error_message_],
                error_delegator_stub_custom_data_example[error_code_],
            )
        elif func_id in [
            func_id_list_host_,
            func_id_list_service_,
        ]:
            # Verify that func is selected and all what is left to do is to query 0...N objects:
            if interp_ctx.curr_container_ipos >= vararg_container_ipos:
                # Search `data_envelope`-s based on existing args on command line:
                vararg_container = interp_ctx.envelope_containers[vararg_container_ipos]
                vararg_container.data_envelopes = (
                    local_server
                    .get_query_engine()
                    .query_data_envelopes_for(vararg_container)
                )

                # Plugin to invoke on client side:
                delegator_plugin_instance_id = self.plugin_instance_id
                # Package into `InvocationInput` payload object:
                invocation_input = InvocationInput.with_interp_context(
                    interp_ctx,
                    delegator_plugin_entry = local_server.plugin_config.plugin_instance_entries[
                        delegator_plugin_instance_id
                    ],
                    custom_plugin_data = {},
                )
                return invocation_input
            else:
                return redirect_to_no_func_error(
                    interp_ctx,
                    local_server.server_config,
                )
        elif func_id in [
            func_id_desc_host_,
            func_id_desc_service_,
        ]:
            # Actual implementation is not defined for demo:
            return redirect_to_error(
                interp_ctx,
                local_server.plugin_config,
                error_delegator_stub_custom_data_example[error_message_],
                error_delegator_stub_custom_data_example[error_code_],
            )
        else:
            # Plugin is given a function name it does not know:
            raise RuntimeError

    @staticmethod
    def invoke_action(invocation_input: InvocationInput):
        """
        Print `data_envelope`-s received from server on client side.
        """

        func_id = get_func_id_from_invocation_input(invocation_input)
        if func_id == func_id_list_host_:
            prohibit_unconsumed_args(invocation_input)
            for data_envelope in invocation_input.envelope_containers[host_container_ipos_].data_envelopes:
                print(json.dumps(data_envelope))
        elif func_id == func_id_list_service_:
            prohibit_unconsumed_args(invocation_input)
            for data_envelope in invocation_input.envelope_containers[service_container_ipos_].data_envelopes:
                print(json.dumps(data_envelope))
        else:
            raise RuntimeError("ERROR: not implemented for demo")



from dataclasses import dataclass, field


@dataclass
class LoadProgressState:
    """
    State and stats to track `data_envelope`-s loading progress.
    """
    envelope_per_col_i: int = field(default = 0)
    envelope_per_col_n: int = field(default = 0)
    total_envelope_i: int = field(default = 0)
    total_envelope_n: int = field(default = 0)



from __future__ import annotations

from copy import deepcopy

import mongomock
from pymongo import MongoClient
from pymongo.collection import Collection
from pymongo.database import Database

from argrelay.misc_helper_common import eprint
from argrelay.mongo_data.LoadProgressState import LoadProgressState
from argrelay.mongo_data.MongoConfig import MongoConfig
from argrelay.runtime_data.EnvelopeCollection import EnvelopeCollection
from argrelay.runtime_data.StaticData import StaticData
from argrelay.schema_config_interp.DataEnvelopeSchema import (
    envelope_id_,
    data_envelope_desc,
    mongo_id_,
)


def get_mongo_client(mongo_config: MongoConfig):
    if mongo_config.use_mongomock:
        return mongomock.MongoClient()
    else:
        return MongoClient(mongo_config.mongo_client.client_connection_string)


def store_envelopes(
    mongo_db: Database,
    cleaned_mongo_collections: set[str],
    static_data: StaticData,
    load_state: LoadProgressState,
):
    # Calculate total:
    for mongo_collection in static_data.envelope_collections:
        envelope_collection: EnvelopeCollection = static_data.envelope_collections[
            mongo_collection
        ]
        load_state.total_envelope_n += len(envelope_collection.data_envelopes)

    # Index all:
    for mongo_collection in static_data.envelope_collections:
        envelope_collection: EnvelopeCollection = static_data.envelope_collections[
            mongo_collection
        ]
        store_envelope_collection(
            mongo_db,
            cleaned_mongo_collections,
            mongo_collection,
            envelope_collection,
            load_state,
        )

    assert load_state.total_envelope_i == load_state.total_envelope_n


def store_envelope_collection(
    mongo_db: Database,
    cleaned_mongo_collections: set[str],
    mongo_collection: str,
    envelope_collection: EnvelopeCollection,
    load_state: LoadProgressState,
) -> None:
    col_proxy: Collection = mongo_db[mongo_collection]
    if mongo_collection not in cleaned_mongo_collections:
        col_proxy.delete_many({})
        col_proxy.drop_indexes()
        cleaned_mongo_collections.add(mongo_collection)

    base_total_envelope_i: int = load_state.total_envelope_i
    load_state.envelope_per_col_i = 0
    load_state.envelope_per_col_n = len(envelope_collection.data_envelopes)
    log_index_progress(mongo_collection, load_state)
    for data_envelope in envelope_collection.data_envelopes:
        if load_state.total_envelope_i > 0 and load_state.total_envelope_i % 1_000 == 0:
            log_index_progress(mongo_collection, load_state)
        load_state.total_envelope_i += 1
        load_state.envelope_per_col_i += 1
        envelope_to_store = deepcopy(data_envelope)

        try:
            data_envelope_desc.validate_dict(envelope_to_store)

            if envelope_id_ in envelope_to_store:
                envelope_to_store[mongo_id_] = data_envelope[envelope_id_]

            col_proxy.insert_one(envelope_to_store)
        except:
            print(f"envelope_to_store: {envelope_to_store}")
            # Rethrow previous error:
            raise

    log_index_progress(mongo_collection, load_state)

    assert load_state.envelope_per_col_i == load_state.total_envelope_i - base_total_envelope_i


def log_index_progress(
    mongo_collection: str,
    load_state: LoadProgressState,
):
    try:
        assert load_state.envelope_per_col_i <= load_state.envelope_per_col_n
        assert load_state.total_envelope_i <= load_state.total_envelope_n
        assert load_state.envelope_per_col_n <= load_state.total_envelope_n
    finally:
        eprint(
            f"collection: {mongo_collection}: indexed envelopes: "
            f"{load_state.envelope_per_col_i}/{load_state.envelope_per_col_n} "
            f"{load_state.total_envelope_i}/{load_state.total_envelope_n} "
            "..."
        )


def log_validation_progress(
    validation_step: str,
    mongo_collection: str,
    load_state: LoadProgressState,
):
    try:
        assert load_state.envelope_per_col_i <= load_state.envelope_per_col_n
        assert load_state.total_envelope_i <= load_state.total_envelope_n
        assert load_state.envelope_per_col_n <= load_state.total_envelope_n
    finally:
        eprint(
            f"validation step: {validation_step}: collection: {mongo_collection}: validated envelopes: "
            f"{load_state.envelope_per_col_i}/{load_state.envelope_per_col_n} "
            f"{load_state.total_envelope_i}/{load_state.total_envelope_n} "
            f"..."
        )


def create_index(
    mongo_db: Database,
    collection_name: str,
    index_props: list[str],
):
    col_proxy: Collection = mongo_db[collection_name]

    for index_prop in index_props:
        col_proxy.create_index(index_prop)



from argrelay.schema_config_plugin.PluginEntrySchema import plugin_entry_desc
from argrelay.schema_request.CallContextSchema import call_context_desc
from argrelay.schema_response.InvocationInputSchema import invocation_input_desc
from argrelay.server_spec.server_data_schema import get_schema_definitions

spec_data = {
    "summary": "Provide input data to delegate execution of command on client side",
    "consumes": [
        "application/json",
    ],
    "produces": [
        "application/json",
    ],
    "parameters": [
        {
            "name": call_context_desc.ref_name,
            "in": "body",
            "required": True,
            "schema": {
                "$ref": "#/definitions/" + call_context_desc.ref_name,
            },
        },
    ],
    "responses": {
        "200": {
            "description": "Input data to delegate execution of command on client side",
            "schema": {
                "$ref": "#/definitions/" + invocation_input_desc.ref_name,
            },
            "examples": {
                "application/json": invocation_input_desc.dict_example,
            },
        },
    },
    "definitions": get_schema_definitions([
        # TODO: provide `call_context_desc` with example of expected comp_type for relay_line_args:
        call_context_desc.ref_name,
        invocation_input_desc.ref_name,
        plugin_entry_desc.ref_name,
    ]),
}



from argrelay.enum_desc.PluginType import PluginType
from argrelay.relay_server.QueryEngine import QueryEngine
from argrelay.runtime_context.AbstractPluginServer import AbstractPluginServer
from argrelay.runtime_data.StaticData import StaticData


class AbstractLoader(AbstractPluginServer):
    """
    `LoaderPlugin` implements loading of `data_envelope` (which are later selected on the CLI via args).

    This is a way to load data once on server start.

    The source of data can be anything (hence, it is a plugin).
    """

    def get_plugin_type(
        self,
    ) -> PluginType:
        return PluginType.LoaderPlugin

    def update_static_data(
        self,
        static_data: StaticData,
        query_engine: QueryEngine,
    ) -> StaticData:
        pass
